# BA EMULATION FUNCTIONS


mkdir = function(workingDirectory, directoryName)
{
	# THIS FUNCTION CREATES NEW DIRECTORIES IN THE GIVEN WORKING DIRECTORY
	localDirs = list.dirs(path = workingDirectory)
	searchString = paste("./", directoryName, sep="")
	directoryExists = searchString %in% localDirs
	if(!directoryExists)
	{
		dir.create(extendPath(workingDirectory, directoryName))
	}
	return(NULL)
}

extendPath = function(path, directoryOrFileName)
{
	# THIS FUNCTION WILL ADD A DIRECTORY OR FILE NAME TO A GIVEN PATH
	chars1 = strsplit(path, "")[[1]]
	chars2 = strsplit(directoryOrFileName, "")[[1]]
	if(chars2[1] == "/")
	{
		directoryOrFileName = paste(chars2[2:length(chars2)], sep="")
	}

	if(chars1[length(chars1)] == "/")
	{
		newPath = paste(path, directoryOrFileName, sep="")
	}
	else
	{
		newPath = paste(path, "/", directoryOrFileName, sep="")
	}
	return(newPath)
}


getCSVHeaderItems = function(path, filename)
{
	# THIS FUNCTION PULLS THE NAMES OF THE COLUMNS FROM A CSV FILE
	fn = extendPath(path, filename)
	con = file(fn)
	header = readLines(con, n = 1)
	colNames = strsplit(header, ",")[[1]]
	return(colNames)
}

getIndicesForZeroColumns = function(path, filename)
{
	fn = extendPath(path, filename)
	df = read.table(fn, sep = ",", header = TRUE)
	zeroCols = c()
	for(i in 1:ncol(df))
	{
		thisCol = df[, i]
		thisCol = thisCol[!is.nan(thisCol) & !is.na(thisCol) & !is.infinite(thisCol)]
		if(min(thisCol) == max(thisCol))
		{
			zeroCols = c(zeroCols, i)
		}
	}
	return(zeroCols)
}


getCSVData = function(path, filename)
{
	# THIS FUNCTION GETS THE DATA FROM A CSV FILE FOR A GIVEN PATH
	chars = strsplit(path, "")
	if(chars[length(chars)] == "/")
	{
		fn = paste(path, filename, sep="")
	}
	else
	{
		fn = paste(path, "/", filename, sep="")
	}
	data = read.table(file = fn, sep=",", header = TRUE)
	return(data)
}

trimWhiteSpaceFromEnds = function(stringVector)
{
	# THIS FUNCTION REMOVE WHITE SPACE FROM THE BEGINNING AND END OF TEXT STRINGS IN A GIVEN VECTOR
	for(i in 1:length(stringVector))
	{
		stringVector[i] = gsub("^\\s+|\\s+$", "", stringVector[i])
	}
	return(stringVector)
}

getTransformsFromTransformationString = function(transformationString)
{
	# THIS FUNCTION TAKES A TRANSFORMATION STRING AND BREAKS IT UP INTO THE SUBSTRINGS THAT ARE SEPARATED BY AN AMPERSAND
	splitString = strsplit(transformationString, "&")[[1]]
	splitString = trimWhiteSpaceFromEnds(splitString)
	return(splitString)
}

applyTransform = function(transformationString, vectorToTransform)
{
	# THIS FUNCTION APPLIES A GIVEN TRANSFORMATION TO A VECTOR OF NUMBERS
	# VALID TRANSFORMATION TYPES ARE LISTED BELOW IN THE FUNCTION DEFINITION
	# NEW TRANSFORMATIONS CAN BE EASILY ADDED
	# MULTIPLE TRANSOFRMATIONS CAN BE APPLIED BY SEPARATING THEN WITH AN AMPERSAND.  THE ORDER IN WHICH TRANSOFRMS ARE THEN APPLIED IS FROM LEFT TO RIGHT.
	transformations = getTransformsFromTransformationString(transformationString)
	if(length(transformations) == 0)
	{
		stop("Error in applyTransform.  The length of the transform string is zero.  Do you mean to use none instead?")
	}
	for(i in 1:length(transformations))
	{
		stringComponents = strsplit(transformations[i], " ")[[1]]
		if(stringComponents[1] == "divideBy")
		{
			vectorToTransform = vectorToTransform/as.numeric(stringComponents[2])
		}
		else if(stringComponents[1] == "cubedRoot")
		{
			vectorToTransform = sign(vectorToTransform)*((abs(vectorToTransform))^(1/3))
		}
		else if(stringComponents[1] == "log")
		{
			vectorToTransform = log(vectorToTransform)
		}
		else if(stringComponents[1] == "log10")
		{
			vectorToTransform = log10(vectorToTransform)
		}
		else if(stringComponents[1] == "none")
		{
			vectorToTransform = vectorToTransform
		}
		else
		{
			stop("Error in applyTransform.  Transformation in string is not defined.")
		}
	}
	return(vectorToTransform)
}


applyReverseTransform = function(transformationString, vectorToTransform)
{
	# THIS FUNCTION REVERSES A GIVEN TRANSFORMATION TO A VECTOR OF NUMBERS
	# VALID TRANSFORMATION TYPES ARE LISTED BELOW IN THE FUNCTION DEFINITION
	# NEW TRANSFORMATIONS CAN BE EASILY ADDED
	# MULTIPLE TRANSOFRMATIONS CAN BE APPLIED BY SEPARATING THEN WITH AN AMPERSAND.  THE ORDER IN WHICH THE REVERSE TRANSOFRMS ARE THEN APPLIED IS FROM RIGHT TO LEFT (OPPOSITE).
	transformations = getTransformsFromTransformationString(transformationString)

	if(length(transformations) == 0)
	{
		stop("Error in applyReverseTransform.  The length of the transform string is zero.  Do you mean to use none instead?")
	}
	transformations = rev(transformations)
	for(i in 1:length(transformations))
	{
		stringComponents = strsplit(transformations[i], " ")[[1]]
		if(stringComponents[1] == "divideBy")
		{
			vectorToTransform = vectorToTransform*as.numeric(stringComponents[2])
		}
		else if(stringComponents[1] == "cubedRoot")
		{
			vectorToTransform = sign(vectorToTransform)*((abs(vectorToTransform))^(3))
		}
		else if(stringComponents[1] == "log")
		{
			vectorToTransform = exp(vectorToTransform)
		}
		else if(stringComponents[1] == "log10")
		{
			vectorToTransform = 10^(vectorToTransform)
		}
		else if(stringComponents[1] == "none")
		{
			vectorToTransform = vectorToTransform
		}
		else
		{
			stop("Error in applyReverseTransform.  Transformation in string is not defined.")
		}
	}
	return(vectorToTransform)
}


getNaturalBoundsForReceptorType = function(receptorName)
{
	# THIS FUNCTION DEFINES THE PLAUSIBLE VALUES THAT DIFFERENT RECEPTOR VARIABLES CAN TAKE.
	stringSplit = strsplit(receptorName, "_")[[1]]
	if(length(stringSplit) < 2)
	{
		stop("Error in getBoundsForReceptorType: receptor name must contain at least two strings separated by an underscore.  The variable type (e.g. dmax) should be the string after the last underscore.")
	}
	variableType = stringSplit[length(stringSplit)]
	if(variableType == "tmax")
	{
		bounds = c(0, Inf)
	}
	else if(variableType == "dmax")
	{
		bounds = c(-Inf, Inf)
	}
	else if(variableType == "Qdr")
	{
		bounds = c(0, Inf)
	}
	else if(variableType == "Qup")
	{
		bounds = c(0, Inf)
	}
	else
	{
		stop("Error in getNaturalBoundsForReceptorType: variable type in receptor name not recognised.")
	}
	return(bounds)
}


getColNumbersFromColNames = function(names, columnNames)
{
	# THIS FUNCTION RETURNS THE COLUMN NUMBERS FOR A GIVEN SET OF NAMES AND COLUMN NAMES
	indices = which(columnNames %in% names)
	return(indices)
}



removeUnwantedColumnNames = function(unwantedNames, columnNames)
{
	# THIS FUNCTION REMOVES UNWANTED COLUMN NAMES FROM A SET OF COLUMN NAMES
	indices = getColNumbersFromColNames(unwantedNames, columnNames)
	return(columnNames[-indices])
}



#' Calculate and visualise (if plotOutFile is given) cross-validation accuracy of model prediction
#' A metrics, similar to ROC area, to density prediction quality using both mean (mu) and standard deviation (sigma)
#' obs, mu, and sigma are vectors of the same length
#' Fast debuging example: obs <<- Y_trans[validationIndices]; mu <<- thisAGP$mean; sigma <<- sqrt(thisAGP$var); 
aveHitRateGaussianDensityForecast <- function(obs, mu, sigma, plotOutFile=NULL){
  confLevels = (seq(0, 0.99999, length.out=100));
  hitRates =rep(0, length(confLevels))
  for(ind in 1:length(confLevels)){
    confLevel = (confLevels[ind] + 1.0)/2.0;
    deviation = qnorm(confLevel);
    hitRates[ind] <- sum(abs((obs-mu)/sigma) <= deviation)    
  }
  hitRates <- hitRates/length(mu)
  aveHitRate = mean(hitRates,na.rm=TRUE)
  if( !is.null(plotOutFile )){
    pdf(paste0(plotOutFile,".pdf"))
    plot(hitRates ~ confLevels,type='l',xlab='1 - confidence level',
         ylab='Hit rate',xlim=c(0,1),ylim=c(0,1),
         col='blue',
         main=sprintf('Average hit rate is %.5f for varying confidence levels', aveHitRate))
    #abline(a=0,b=1, col='blue',lty= "dotted")
    lines(c(0,1),c(0,1),col='blue',lty='dashed')
    abline(a=1,b=0, col='grey60',lty='dotted')
    dev.off();
  }
  return(aveHitRate)  
}



buildEmulators = function(	workingDirectory, 
							parameterNames, 
							parameterTransforms, 
							receptorNames, 
							receptorTransforms, 
							receptorRawMin, 
							receptorRawMax, 
							designFileName, 
							receptorFileName, 
							csvFileDir, 
							parallel = FALSE, 
							appendSuffix = "",
							paramMin, 
							paramMax,
							createEmulatorPerformancePlots)
{
	# THIS FUNCTION BUILDS A SET OF EMULATORS FOR A GIVEN PARAMETER MATRIX AND VECTORS OF RECEPTORS
	mkdir(workingDirectory, "emulators")
	
	# THE EMULATORS GET BUILT WITH TRANSFORMED PARAMETERS AND OUTPUTS
	parameterMatrix = read.table(extendPath(csvFileDir, designFileName), header = TRUE, sep = ",", comment.char="")
	parameterMatrix = parameterMatrix[, parameterNames]
	parameterMatrix_trans = parameterMatrix
	for(i in 1:ncol(parameterMatrix_trans))
	{
		parameterMatrix_trans[, i] = applyTransform(parameterTransforms[i], parameterMatrix_trans[, i])
	}
	
	# Transform the output from the model if appropriate
	receptorMatrix = read.table(extendPath(csvFileDir, receptorFileName), header = TRUE, sep = ",", comment.char="")
	colnames(receptorMatrix) = paste(colnames(receptorMatrix), appendSuffix, sep="")
	receptorMatrix = receptorMatrix[, receptorNames]
	receptorMatrix_trans = receptorMatrix
	for(i in 1:ncol(receptorMatrix_trans))
	{
		receptorMatrix_trans[, i] = applyTransform(receptorTransforms[i], receptorMatrix_trans[, i])
	}
	if(parallel)
	{
		foreach(i = seq(1, ncol(receptorMatrix_trans), 1)) %dopar% {
			buildAGPEmulator(parameterMatrix = 	parameterMatrix_trans,
												receptorVals = receptorMatrix[, i], 
												transformedReceptorVals = receptorMatrix_trans[, i], 
												receptorTransform = receptorTransforms[i], 
												receptorRawMin = receptorRawMin[i], 
												receptorRawMax = receptorRawMax[i], 
												emulatorDirectory = extendPath(workingDirectory, "emulators"), 
												receptorName = receptorNames[i],
												paramMin, 
												paramMax,
												createEmulatorPerformancePlots)
		}
	}
	else
	{
		for(i in seq(176, ncol(receptorMatrix_trans), 1))
		{
			buildAGPEmulator(parameterMatrix = parameterMatrix_trans, receptorVals = receptorMatrix[, i], transformedReceptorVals = receptorMatrix_trans[, i], receptorTransform = receptorTransforms[i], receptorRawMin = receptorRawMin[i], receptorRawMax = receptorRawMax[i], emulatorDirectory = extendPath(workingDirectory, "emulators"), receptorName = receptorNames[i])
		}
	}
	
	return(NULL)
}





getEmulatorMaxVal = function(	parameterMatrix,
								transformedReceptorVals, 
								startGlobal, 
								endGlobal,
								numValidationPoints)
{
	randSeq = sample(seq(1, nrow(parameterMatrix)))
	blockInd = 1:numValidationPoints
	paramVals = (startGlobal + 1):endGlobal #WJ tried to save a bit search time seq((startGlobal + 1),endGlobal,by=3)#
	checkVals = rep(Inf, length(paramVals))
	ok = FALSE
	for(i in 1:length(paramVals))
	{
		coverage95 = c()
		coverage50 = c()
		validationIndices = randSeq[blockInd]
		buildIndices = randSeq[-blockInd]
		if(i > length(paramVals))
		{
			return(endGlobal)
		}
		
		thisAGP = aGP(X = parameterMatrix[buildIndices, ], Z = transformedReceptorVals[buildIndices], XX = parameterMatrix[validationIndices, ], 
                  start = startGlobal, end = paramVals[i], verb = 0,
                  method='mspe')
        validationMeans = thisAGP$mean
		validationLower = thisAGP$mean - 1.96*sqrt(thisAGP$var)
		validationUpper = thisAGP$mean + 1.96*sqrt(thisAGP$var)
		
		validationLower2 = thisAGP$mean - 0.6745*sqrt(thisAGP$var)
		validationUpper2 = thisAGP$mean + 0.6745*sqrt(thisAGP$var)
		
		theseTransformedReceptorVals = transformedReceptorVals[validationIndices]
		
		# SOMETIMES THE EMULATOR FAILS TO PREDICT AND GIVES NAN 
		failedIndices = which(is.nan(validationMeans) | is.infinite(validationMeans) | is.na(validationMeans))
		if(length(failedIndices) > 0)
		{
			validationMeans = validationMeans[-failedIndices]
			validationSD = sqrt(thisAGP$var[-failedIndices])
			validationLower = validationLower[-failedIndices]
			validationUpper = validationUpper[-failedIndices]
			validationLower2 = validationLower[-failedIndices]
			validationUpper2 = validationUpper[-failedIndices]
			validationIndices = validationIndices[-failedIndices]
			theseTransformedReceptorVals = theseTransformedReceptorVals[-failedIndices]
		}
		else
		{
			validationSD = sqrt(thisAGP$var)
		}
		
		minY = min(validationLower)
		maxY = max(validationUpper)
		
		residuals = c(residuals, (theseTransformedReceptorVals - validationMeans))
		
		theseHits95 = ifelse(((theseTransformedReceptorVals >= validationLower) &  (theseTransformedReceptorVals <= validationUpper)), 1, 0)
		theseHits50 = ifelse(((theseTransformedReceptorVals >= validationLower2) &  (theseTransformedReceptorVals <= validationUpper2)), 1, 0)
		
		coverage95 = c(coverage95,  theseHits95)
		coverage50 = c(coverage50,  theseHits50)
		
		hr95 = sum(coverage95)/length(coverage95)
		hr50 = sum(coverage50)/length(coverage50)
		
		checkVals[i] = abs(log(hr95/(1 - hr95)) - log(0.95/(1 - 0.95))) + abs(log(hr50/(1 - hr50)) - log(0.5/(1 - 0.5)))
	}
	ind = which(checkVals == min(checkVals))
	ind = ind[1]
	return(paramVals[ind])
}





buildAGPEmulator = function(parameterMatrix, 
							receptorVals, 
							transformedReceptorVals, 
							receptorTransform, 
							receptorRawMin, 
							receptorRawMax, 
							emulatorDirectory, 
							csvDirectory, 
							receptorName, 
							paramMin, 
							paramMax, 
							createEmulatorPerformancePlots = FALSE)
{
	# THIS FUNCTION IS CALLED BY BUILDEMULATORS AND CREATES APPROXIMATE LOCAL GAUSSIAN PROCESS EMULATORS
	mkdir(emulatorDirectory, receptorName)
	
	#Check that parameterMatrix and receptorVals are the same length.  If not shorten parameterMatrix to the correct length
	if(nrow(parameterMatrix) > length(receptorVals))
	{
		print("Truncating parameter matrix.  You have less receptors.  Did you not use the full DoE?")
		parameterMatrix = parameterMatrix[1:length(receptorVals), ]
	}
	else if(nrow(parameterMatrix) < length(receptorVals))
	{
		print("Error: You have more receptor values that rows in the parameter matrix")
	}
	
	# BEFORE CREATING THE EMULATOR, REMOVE ANY ENTRIES FROM receptorVals THAT ARE NAN, NA OR INF
	# ALSO NEED TO REMOVE THE CORRESPONDING ROWS FROM THE PARAMETER MATRIX
	badIndices1 = which(is.nan(transformedReceptorVals) | is.infinite(transformedReceptorVals) | is.na(transformedReceptorVals))
	badIndices2 = which(receptorVals < receptorRawMin | receptorVals > receptorRawMax)
	badIndices = unique(c(badIndices1, badIndices2))
  if(length(badIndices)!=0)
  {
	  transformedReceptorVals = transformedReceptorVals[-badIndices]
	  parameterMatrix = parameterMatrix[-badIndices, ]
  }
  #Rescale parameters to lie between 0 and 1
  for(i in 1:ncol(parameterMatrix))
  {
  	parameterMatrix[,i] = (parameterMatrix[,i] - paramMin[i])/(paramMax[i] - paramMin[i])
  }
  
	validationFoldSize = 30
	#Perform an n-fold cross validation (using the parameter validationFoldSize, default value is 30)
	randSeq = sample(seq(1, nrow(parameterMatrix)))
	samplesPerFold = floor(nrow(parameterMatrix)/validationFoldSize) 
	totalValidationPoints = samplesPerFold*validationFoldSize
	
	thisEmulatorMaxVal = getEmulatorMaxVal(parameterMatrix, transformedReceptorVals, startGlobal, endGlobal, round(length(row.names(parameterMatrix))/10))
	
	hitRate = rep(NA, validationFoldSize)
    CRPSs = aveHitRates = rep(NA, validationFoldSize)
    cat(receptorName, "\n")
    coverage95 = c()
    coverage50 = c()
    residuals = c()

	for(i in 1:validationFoldSize)
	{
		#Get the validation data together and build the aGP emulator
		blockInd = ((i - 1)*samplesPerFold + 1):(i*samplesPerFold)
		validationIndices = randSeq[blockInd]
		buildIndices = randSeq[-blockInd]
		#d <- darg(list(mle = TRUE), parameterMatrix[buildIndices, ])
		#g <- garg(list(mle = TRUE), transformedReceptorVals[buildIndices])
		#print(range(parameterMatrix[buildIndices, ]))
		#print(range(transformedReceptorVals[buildIndices]))
		thisAGP = aGP(X = parameterMatrix[buildIndices, ], Z = transformedReceptorVals[buildIndices], XX = parameterMatrix[validationIndices, ], 
                  start = startGlobal, end = thisEmulatorMaxVal, verb = 0,
                  method='mspe')
                  
		validationMeans = thisAGP$mean
		validationLower = thisAGP$mean - 1.96*sqrt(thisAGP$var)
		validationUpper = thisAGP$mean + 1.96*sqrt(thisAGP$var)
		
		validationLower2 = thisAGP$mean - 0.6745*sqrt(thisAGP$var)
		validationUpper2 = thisAGP$mean + 0.6745*sqrt(thisAGP$var)
		
		theseTransformedReceptorVals = transformedReceptorVals[validationIndices]
		
		# SOMETIMES THE EMULATOR FAILS TO PREDICT AND GIVES NAN 
		failedIndices = which(is.nan(validationMeans) | is.infinite(validationMeans) | is.na(validationMeans))
		if(length(failedIndices) > 0)
		{
			validationMeans = validationMeans[-failedIndices]
			validationSD = sqrt(thisAGP$var[-failedIndices])
			validationLower = validationLower[-failedIndices]
			validationUpper = validationUpper[-failedIndices]
			validationLower2 = validationLower[-failedIndices]
			validationUpper2 = validationUpper[-failedIndices]
			validationIndices = validationIndices[-failedIndices]
			theseTransformedReceptorVals = theseTransformedReceptorVals[-failedIndices]
		}
		else
		{
			validationSD = sqrt(thisAGP$var)
		}
		
		minY = min(validationLower)
		maxY = max(validationUpper)
		
		residuals = c(residuals, (theseTransformedReceptorVals - validationMeans))
		
		theseHits95 = ifelse(((theseTransformedReceptorVals >= validationLower) &  (theseTransformedReceptorVals <= validationUpper)), 1, 0)
		theseHits50 = ifelse(((theseTransformedReceptorVals >= validationLower2) &  (theseTransformedReceptorVals <= validationUpper2)), 1, 0)
		
		coverage95 = c(coverage95,  theseHits95)
		coverage50 = c(coverage50,  theseHits50)
		
		if(createEmulatorPerformancePlots)
		{
			#create plots to visualise each of the cross validation folds
			#WJ: as well as the hit rate for different confidence level
			aveHitRates[i] <- aveHitRateGaussianDensityForecast(transformedReceptorVals[validationIndices], validationMeans, validationSD,
															plotOutFile = extendPath(extendPath(emulatorDirectory, receptorName), paste(receptorName, "_hitRates_fold_", i, ".pdf", sep="") )  )
		
			CRPSs[i] <- crps(theseTransformedReceptorVals, data.frame(validationMeans, validationSD))$CRPS;
			pdf(file = extendPath(extendPath(emulatorDirectory, receptorName), paste(receptorName, "_fold_", i, ".pdf", sep="") ) )
			plot(theseTransformedReceptorVals, validationMeans, xlab = paste("Model output ", "(transform =", receptorTransform, ")"),
			 ylab = paste(receptorName, " (transform =", receptorTransform, ")"
			), main = paste("laGP Model Emulation for Model Output: ", receptorName), ylim = c(minY, maxY))
			for(j in 1:length(validationMeans))
			{
				lines(rep(theseTransformedReceptorVals[j], 2), c(validationLower[j], validationUpper[j]), col="gray" )
			}
			points(theseTransformedReceptorVals, validationMeans)
			abline(0,1, col="red", lty=2)
			
			coveredIndices = which(theseTransformedReceptorVals > validationLower & theseTransformedReceptorVals < validationUpper)
			hitRate[i] = length(coveredIndices)/length(validationIndices)*100
	
			#WJ: output oneMetrics as well as CRPS
			#cat(sprintf("Average Hit Rate for output: %.5f; CRPS: %.5f\n", aveHitRates[i], CRPSs[i]))
			#cat("_________________________________________________________________ \n")
			text(mean(range(validationMeans)), maxY, sprintf('hitRate=%.4f (0.05 level), mean hit pct=%.4f; CRPS=%.5f', 
											   hitRate[i]/100, aveHitRates[i], CRPSs[i]))
			dev.off()
		}
		
	}
	colNames = c("EmulatorName", "MSE", "MAE", "Bias", "SD", "Q0.05", "Q0.25", "Q0.5", "Q0.75", "Q0.95", "Coverage95", "Coverage50", "FoldSize", "globalEnd", "HitRate50", "HitRate95", "EmulatorIsOK")
	
	thisMSE = mean(residuals^2)
	thisMAE = mean(abs(residuals))
	thisBias = mean(residuals)
	thisSD = sd(residuals)
	theseQuantiles = quantile(residuals, c(0.05, 0.25, 0.5, 0.75, 0.95))

	thisCoverage95 = sum(coverage95)
	thisCoverage50 = sum(coverage50)
	thisNumTrials = length(residuals)
	theseSummaryStats = c(receptorName, thisMSE, thisMAE, thisBias, thisSD, theseQuantiles, thisCoverage95, thisCoverage50, thisNumTrials, thisEmulatorMaxVal)
	
	hr95 = thisCoverage95/length(coverage95)
	hr50 = thisCoverage50/length(coverage50)
	theseSummaryStats = c(theseSummaryStats, hr50, hr95)
	
	#CREATE THESE INTERVALS FOR MAKING A DECISION
	#EACH EMULATORS COVERAGE IS A BERNOULLI TRIAL, WITH A PROBABILITY OF 0.95 (or 0.5) of COVERING THE MODELED VALUE
	#WE COMPUTE THE INTERVAL THAT THE EMULATOR COVERAGE SHOULD FALL WITHIN 999 times out of 1000
	#WE THEREFORE EXPECT A GOOD EMULATOR TO BE DISCARDED WITH A PROBABILITY OF 0.001  
	#CI95_lower =qbinom(0.0005, length(coverage95), 0.95)
	#CI50_lower = qbinom(0.0005, length(coverage50), 0.5)
	#CI95_upper = qbinom(0.9995, length(coverage95), 0.95)
	#CI50_upper = qbinom(0.9995, length(coverage50), 0.5)
  
	#Considering the hitRates are from cross-validation results, we can loose the 
	CI95_lower = 0.9*length(coverage95) 
	CI50_lower = 0.25*length(coverage95)
	CI95_upper = 0.995*length(coverage95) 
	CI50_upper = 0.75*length(coverage95)
	
	
	if(thisCoverage95 >= CI95_lower & thisCoverage50 >= CI50_lower  &  thisCoverage95 <= CI95_upper  &  thisCoverage50 <= CI50_upper)
	{
		theseSummaryStats = c(theseSummaryStats, 1)
	}
	else
	{
		theseSummaryStats = c(theseSummaryStats, 0)
	}

	

	thisLine = t(theseSummaryStats)
	colnames(thisLine) = colNames
	if(file.exists(extendPath(csvDirectory, "emulatorSummaryStatistics.csv")))
	{
		write.table(x = thisLine, file = extendPath(csvDirectory, "emulatorSummaryStatistics.csv"), append = TRUE, sep=",", quote = FALSE, col.names = FALSE, row.names = FALSE)
	}
	else
	{
		write.table(x = thisLine, file = extendPath(csvDirectory, "emulatorSummaryStatistics.csv"), append = TRUE, sep=",", quote = FALSE, col.names = TRUE, row.names = FALSE)
	}
	
	
	
	if(createEmulatorPerformancePlots)
	{
		#Create plots to visualise the hit rates   
		pdf(file = extendPath(extendPath(emulatorDirectory, receptorName), paste("hitRates", ".pdf", sep="") ) )
	
		plot(hitRate, xlab = "Cross-Validation Fold Index", ylab = "Hit Rate (Percent)",
		   main=sprintf('statistics of hit rates: mean= %.4f; sd=%.4f', mean(hitRate), sd(hitRate)))
		abline(95, 0, col="red")
		dev.off()
		#WJ: Create plots to visualise the hit rates   
		pdf(file = extendPath(extendPath(emulatorDirectory, receptorName), paste("averageHitRates", ".pdf", sep="")))
	
		plot(aveHitRates, xlab = "Cross-Validation Fold Index", ylab = "Hit Rate (Percent)",
			 main=sprintf('statistics of hit rates: mean= %.4f; sd=%.4f', mean(aveHitRates), sd(aveHitRates)))
		abline(95, 0, col="red")
		dev.off()
	}
	
	
	
	#cat(sprintf("Summary: Average Hit Rate 4 folds: %.5f (0.05level); %.5f (all). CRPS %.5f \n", mean(hitRate)/100, mean(aveHitRates), mean(CRPSs)))
    
	thisAGP = list(X = parameterMatrix, Z = transformedReceptorVals, endGlobal = thisEmulatorMaxVal)
	
	saveRDS(object = thisAGP, file = extendPath(extendPath(emulatorDirectory, receptorName), paste(receptorName, ".RData", sep="")) )
	return(NULL)
}




logPrior = function(paramVector)
{
	# THIS FUNCTION DEFINES THE LOG PRIOR FOR A VECTOR OF PARAMETERS AND IS USED BY THE MCMC FUNCTION (PART OF ADAPTMCMC PACKAGE)
	if(length(paramVector) != length(mcmcParams))
	{
		stop("Error in logPrior: mcmcParams is of different length to paramVector.")
	}
	# This function makes use of global variables (a list) called "mcmcParams" which must be defined before running this function
	logPrior = 0
	dimCounter = 0
	for(i in 1:length(mcmcParams))
	{
		thisDim = length(mcmcParams[[i]]$transform)
		calibrationParamPrior = mcmcParams[[i]]$priorType
		thisParamVectorSegment = paramVector[(dimCounter + 1):(dimCounter + thisDim)]
		if(any(thisParamVectorSegment < mcmcParams[[i]]$min) | any(thisParamVectorSegment > mcmcParams[[i]]$max))
		{
			# We are outside the allowable bounds of the parameters
			return(-Inf)
		}
		else if(calibrationParamPrior == "normal")
		{
			logPrior = logPrior + dnorm(thisParamVectorSegment, mcmcParams[[i]]$mean, mcmcParams[[i]]$sd, log = TRUE)
		}
		else if(calibrationParamPrior == "weibull")
		{
			logPrior = logPrior + dweibull(thisParamVectorSegment, mcmcParams[[i]]$shape, mcmcParams[[i]]$scale, log = TRUE)
		}
		else if(calibrationParamPrior == "multivariateNormal")
		{
			logPrior = logPrior + dmvnorm(thisParamVectorSegment, mcmcParams[[i]]$mu, mcmcParams[[i]]$cov, log = TRUE)
		}
		else
		{
			stop("Error in logPrior: Undefined prior distribution in function.  You may need to modify the code for your prior.")
		}
		dimCounter = dimCounter + thisDim
	}
	return(logPrior)
}


logPosterior = function(paramVector)
{
	# THIS FUNCTION DEFINES THE LOG POSTERIOR FOR A VECTOR OF PARAMETERS AND IS USED BY THE MCMC FUNCTION (PART OF ADAPTMCMC PACKAGE)
	# This function makes use of the global variables (list) mcmcParams which must be defined beforehand
	# and also mcmcObservationError
	
	logLikelihood = 0
	for(i in (1:length(mcmcObservationError))){
		emulatedValues = laGP(Xref = paramVector, X = mcmcObservationError[[i]]$emulator$X, Z = mcmcObservationError[[i]]$emulator$Z, end = endGlobal, start = startGlobal, verb = 0, method='alc')
		if(!is.nan(emulatedValues$mean))
		{
			modelOutput = sqrt(emulatedValues$s2)*rt(1, emulatedValues$df) + emulatedValues$mean
			modelOutput = applyReverseTransform(mcmcObservationError[[i]]$outputTransform, modelOutput)
			logLikelihood = logLikelihood + dnorm(modelOutput, mcmcObservationError[[i]]$params$mean, mcmcObservationError[[i]]$params$sd, log = TRUE)
		}
		else
		{
			return(-Inf)
		}

	}
	
	lp = logPrior(paramVector)
	logPosterior = lp + logLikelihood
  print(logPosterior)
	return(logPosterior)
}


constructCalibrationEmulatorList = function(emulatorNames, emulatorOutputTransforms, outputDistributions, outputParamsList, outputMins, outputMaxs)
{
	# THIS FUNCTION CONSTRUCTS A LIST OF EMULATORS FOR USE IN CALIBRATING A MODEL WITH THE MCMC FUNCTION
	emulatorList = list()
	for(i in 1:length(emulatorNames))
	{
		thisEmulator = readRDS(extendPath(extendPath(extendPath(getwd(), "emulators"), emulatorNames[i]), paste(emulatorNames[i], ".RData", sep="")))
		emulatorList[[emulatorNames[i]]] = list(emulator = thisEmulator, outputTransform = emulatorOutputTransforms[i], outputDistribution = outputDistributions[i], params = outputParamsList[[i]], min = outputMins[i], max = outputMaxs[i])
	}
	return(emulatorList)
}



samplePriors = function(numSamples, parameterList)
{
	# THIS FUNCTION SAMPLES FROM THE PRIOR DISTRIBUTIONS DEFINED IN THE PARAMETER LIST
	for(i in 1:length(parameterList))
	{
		thisSampledVector = rep(NA, numSamples)
		if(parameterList[[i]]$priorType == "multivariateNormal")
		{
			counter = 0
			while(counter < numSamples)
			{
				cholCov = chol(parameterList[[i]]$cov)
				mu = parameterList[[i]]$mu
				vec = t(as.matrix(mu)) + cholCov %*% as.matrix(rnorm(length(mu)))
				if(all(vec >= parameterList[[i]]$min) & all(vec <= parameterList[[i]]$max))
				{
					if(counter == 0)
					{
						thisSampledVector = t(vec)
					}
					else
					{
						thisSampledVector = rbind(thisSampledVector, t(vec))
					}
					counter = counter + 1
				}
			}
			colnames(thisSampledVector) = parameterList[[i]]$paramName
		}
		else if(parameterList[[i]]$priorType == "normal")
		{
			counter = 0
			while(counter < numSamples)
			{
				vec = as.matrix(rnorm(1, parameterList[[i]]$mean, parameterList[[i]]$sd))
				if(all(vec >= parameterList[[i]]$min) & all(vec <= parameterList[[i]]$max))
				{
					if(counter == 0)
					{
						thisSampledVector = vec
					}
					else
					{
						thisSampledVector = rbind(thisSampledVector, vec)
					}
					counter = counter + 1
				}
			}
			colnames(thisSampledVector) = parameterList[[i]]$paramName
		}
		else if(parameterList[[i]]$priorType == "weibull")
		{
			counter = 0
			while(counter < numSamples)
			{
				vec = as.matrix(rweibull(1, parameterList[[i]]$shape, parameterList[[i]]$scale))
				if(all(vec >= parameterList[[i]]$min) & all(vec <= parameterList[[i]]$max))
				{
					if(counter == 0)
					{
						thisSampledVector = vec
					}
					else
					{
						thisSampledVector = rbind(thisSampledVector, vec)
					}
					counter = counter + 1
				}
			}
			colnames(thisSampledVector) = parameterList[[i]]$paramName
		}
		else if(parameterList[[i]]$priorType == "uniform")
		{
			vec = as.matrix(runif(numSamples, parameterList[[i]]$min, parameterList[[i]]$max))
			thisSampledVector = vec
			colnames(thisSampledVector) = parameterList[[i]]$paramName
		}
		if(i == 1)
		{
			samples = thisSampledVector
		}
		else
		{
			samples = cbind(samples, thisSampledVector)
		}
	}
	return(samples)
}


sampleFromEmulators = function(workingDirectory, 
                               parameterNames, 
                               receptorNames, 
                               receptorTransforms, 
                               parameterSamples, 
                               receptorMin=-Inf,
                               receptorMax=Inf,
                               useReseptorBoundsFromDOE=TRUE,
                               parameterMin,
                               parameterMax,
                               parallel)
{
	emulatorDirectory = extendPath(workingDirectory, "emulators")
	if(parallel)
	{
		foreach(i = seq(1, length(receptorNames), 1)) %dopar% {
		  print(receptorNames[i])
			thisAGP = readRDS(file = extendPath(extendPath(emulatorDirectory, receptorNames[i]), paste(receptorNames[i], ".RData", sep="")) )
			writeEmulatorSamples(emulatorDirectory = emulatorDirectory, emulator = thisAGP, parameterSamples = parameterSamples, receptorTransform = receptorTransforms[i], receptorName=receptorNames[i])
		}
	}
	else
	{
		for(i in 1:length(receptorNames))
		{
		  print(receptorNames[i])
      thisAGP = readRDS(file = extendPath(extendPath(emulatorDirectory, receptorNames[i]), paste(receptorNames[i], ".RData", sep="")) )
			writeEmulatorSamples(emulatorDirectory = emulatorDirectory, 
                           emulator = thisAGP, 
                           parameterSamples = parameterSamples, 
                           receptorTransform = receptorTransforms[i], 
                           receptorName=receptorNames[i],
                           receptorMin = receptorMin[i], 
                           receptorMax = receptorMax[i],
                           parameterMin = parameterMin,
                           parameterMax = parameterMax)
		}
	}
	return(NULL)
}

writeEmulatorSamples = function(emulatorDirectory, 
                                emulator, 
                                parameterSamples, 
                                receptorTransform, 
                                receptorName,
                                receptorMin,
                                receptorMax,
                                parameterMin,
                                parameterMax,
                                useReseptorBoundsFromDOE=TRUE)
{
	if(useReseptorBoundsFromDOE){receptorMin=min(emulator$Z)}
  if(useReseptorBoundsFromDOE){receptorMax=max(emulator$Z)}
  n = nrow(parameterSamples)
	#sampleDirectory = extendPath(workingDirectory, "samples")
  modelOutput = rep(NA, n)
  # Rescale all parameters to between 0 and 1 before using the emulator
  for(i in ncol(parameterSamples))
  {
  	parameterSamples = (parameterSamples - parameterMin[i])/(parameterMax[i] - parameterMin[i])
  }
	for(i in 1:n)
	{
		#d <- darg(list(mle = TRUE), emulator$X)
		#g <- garg(list(mle = TRUE), emulator$Z)
		emulatedValues = laGP(Xref = parameterSamples[i, ], 
                          X = emulator$X, 
                          Z = emulator$Z, 
                          end = endGlobal, 
                          start = startGlobal, 
                          verb = 0, 
                          method='mspe')
		if(!is.nan(emulatedValues$mean))
		{
		  modelOutput[i] = -Inf
      teller = 0
		  while( (modelOutput[i]<receptorMin | modelOutput[i]>receptorMax) & teller < 1000 )
		  {
		    modelOutput[i] = emulatedValues$mean + sqrt(emulatedValues$s2)*rt(1, emulatedValues$df)
        	teller = teller + 1
		  }
		  if(teller == 1000 & (modelOutput[i]<receptorMin | modelOutput[i]>receptorMax))
		  {
		  	modelOutput[i] = NA
		  }
    }
		else
		{
			modelOutput[i] = NA
		}
	}
	emulatedOutputs = applyReverseTransform(receptorTransform, modelOutput)
	write.table(emulatedOutputs, file = extendPath(emulatorDirectory, paste0(receptorName,"_prediction.csv")), sep=",", row.names = FALSE)
}